import 'dart:async';
import 'dart:developer';
import 'dart:io';

import 'package:generator_plus/src/generation/base.dart';
import 'package:generator_plus/src/generation/file_processor.dart';
import 'package:generator_plus/src/generation/new_file.dart';
import 'package:generator_plus/src/generators/utils/template_utils.dart';
import 'package:generator_plus/src/generators/utils/variables_resolver.dart';

import './interactions/base.dart';
import 'package:generator_plus/src/generator_base.dart';
import 'package:generator_plus/src/generators/base.dart';
import 'package:generator_plus/src/common.dart';
import 'package:logging/logging.dart';
import "package:path/path.dart" as path;
import 'package:args/command_runner.dart';
import 'package:mustache/mustache.dart';
import 'package:console/console.dart';

/// A class in charge of running a [GeneratorPlus] implementation in a fully
/// compatible manner with a CLI interface
class CLIGeneratorPlusRunner extends GeneratorPlusRunnerAPI {
  @override
  CLInterface clInterface;
  CommandRunner commandRunner;
  Logger logger;
  VariablesResolver varsResolver;

  CLIGeneratorPlusRunner(
      GeneratorPlus implementation,
      this.clInterface,
      CLIInteractionsHandler interactionsHandler,
      this.logger,
      this.varsResolver)
      : super(implementation, clInterface, interactionsHandler, logger,
            varsResolver) {
    commandRunner = new CommandRunner("generator-plus",
        "An automatic metaprogramming tool for faster, easier and happier :-D development");
    //CLI setup
    commandRunner.argParser.addFlag("pub-get",
        help: "Executes 'pub get' after the execution of the generator",
        negatable: false);
    commandRunner.argParser.addFlag("quiet",
        help: "Do not output any workflow log (not even the given by verbose)",
        abbr: 'q');
    commandRunner.argParser.addOption("verbose",
        help: "Outputs execution logging messages to stdout",
        abbr: 'v',
        allowedHelp: {
          '1': 'fine verbose level',
          '2': 'finer verbose level',
          '3': 'finest verbose level'
        });
    for (Generator generator in implementation.generators) {
      commandRunner.addCommand(new GeneratorPlusCommand(generator));
    }
  }

  @override
  Future printUsage() async {
    String pathResolved =
        await resolvePath("package:generator_plus/ascii_art.txt");
    File asciiArt = new File(pathResolved);
    String str = asciiArt.readAsStringSync();
    TextPen textPen = new TextPen();
    textPen.blue();
    textPen(str);
    textPen.normal();
    Console.setBold(true);
    clInterface.print(textPen.buffer.toString());
    Console.setBold(false);
    clInterface.println(commandRunner.usage);
  }

  @override
  Future execution(List<String> args) {
    if (args.isEmpty || args.last.endsWith("help") || args.last == "-h")
      return printUsage();
    return commandRunner.run(args).catchError((e) {
      if (e is UsageException || e is TemplateException)
        iface.printError(e.message);
      else
        iface.printError(e.toString());
      throw e;
    });
  }
}

class GeneratorPlusCommand extends Command {
  Generator generator;

  GeneratorPlusCommand(this.generator);

  @override
  String get name => generator.name;

  @override
  String get description => generator.description;

  Future run() {
    // globalResults -> todos los argumentos de la CLI
    // argResults -> los argumentos del comando nomas
    bool runPubGet =
        globalResults.wasParsed("pub-get") ? globalResults["pub-get"] : false;
    if (globalResults.wasParsed("verbose")) {
      Level level;
      switch (globalResults["verbose"]) {
        case '1':
          level = Level.FINE;
          break;
        case '2':
          level = Level.FINER;
          break;
        case '3':
          level = Level.FINEST;
          break;
      }
      generator.logger.level = level;
    }
    if (!(globalResults.wasParsed("quiet") && globalResults["quiet"])) {
      TextPen pen;
      generator.logger.onRecord.listen((LogRecord rec) {
        pen = new TextPen();
        bool supressStandardOutput = false;
        if (rec.object is GenerationResult) {
          if (rec.object is FileGenerationResult) {
            FileGenerationResult res = rec.object;

            pen.green();
            pen("Created file '${path.dirname(res.object.path)}/\x1b[1m");
            pen.lime();
            pen(path.basename(res.object.path));
            pen.green();
            pen("\x1b[0m'");
            supressStandardOutput = true;
          } else if (rec.object.runtimeType == FileProcessResult) {
            FileProcessResult res = rec.object;
            pen.gold();
            pen("Processed file '${path.dirname(res.object.fileChanged.path)}/\x1b[1m");
            pen.yellow();
            pen(path.basename(res.object.fileChanged.path));
            pen.gold();
            pen("\x1b[0m'\n");
            pen.green();
            res.object.linesAdded.forEach((int lineNum, String line) {
              pen("    + $lineNum:  $line\n");
            });
            res.object.linesChanged.forEach((int lineNum, LineChange line) {
              pen.gold();
              pen("    - $lineNum:  ${line.before}\n");
              pen.yellow();
              pen("    + $lineNum:  ${line.after}\n");
            });
            pen.red();
            res.object.linesRemoved.forEach((int lineNum, String line) {
              pen("    - $lineNum:  $line\n");
            });
            supressStandardOutput = true;
          }
        }
        if (rec.level == Level.SEVERE) pen.darkRed();
        if (rec.level == Level.WARNING) pen.gold();
        if (rec.level == Level.CONFIG) pen.yellow();
        if (rec.level == Level.INFO) pen.cyan();
        if (rec.level == Level.FINE) pen.white();
        if (rec.level == Level.FINER) pen.lightGray();
        if (rec.level == Level.FINEST) pen.gray();
        if (!supressStandardOutput) pen('${rec.time}: ${rec.message}');
        pen.normal();
        pen();
      });
    }
    return generator.execute(runPubGetDependencies: runPubGet);
    // generator.execute(runPubGetDependencies: runPubGet).then((r) {
    //   print("andó??");
    // });
  }
}
