import 'dart:async';
import 'dart:io';
import 'package:generator_plus/src/generators/utils/variables_resolver.dart';
import 'package:mustache/mustache.dart';
import 'package:mustache_recase/mustache_recase.dart' as mustache_recase;
import "package:path/path.dart" as path;

String original_package_path;

Template partialsResolver(String partialName) {
  File file = new File(
      "${getPackageRootPath()}lib/src/mustache_partials/$partialName.mustache");
  if (!file.existsSync()) {
    if (original_package_path == null)
      throw new MissingPartialException(partialName, false);
    else
      file = new File(path.normalize(
          "$original_package_path/lib/src/mustache_partials/$partialName.mustache"));
  }
  if (file.existsSync())
    return new Template(file.readAsStringSync(),
        partialResolver: partialsResolver);
  else
    throw new MissingPartialException(partialName, true);
}

class MissingPartialException implements Exception {
  final String partialName;
  final bool original_package_searched;
  MissingPartialException(this.partialName, this.original_package_searched);

  String toString() =>
      "Missing partial: Partial '$partialName' not found " +
      (original_package_searched
          ? "neither on ${getPackageRootPath()}lib/src/mustache_partials/$partialName.mustache nor package:generator_plus/src/mustache_partials/$partialName.mustache"
          : "on ${getPackageRootPath()}lib/src/mustache_partials/$partialName.mustache");
}

String getThisPackageName() {
  Directory dir = getPackageRoot();
  File pubspec =
      dir.listSync().singleWhere((f) => f.path.endsWith("pubspec.yaml"));
  RegExp regExp = new RegExp("name: [\"']?([\\w-]*)[\"']?");
  String ret = regExp.firstMatch(pubspec.readAsStringSync())?.group(1);
  if(ret == null || ret == "")
    throw new Error();
  return ret;
}

Directory getPackageRoot() {
  Directory dir = Directory.current;
  try {
    while (
        dir.listSync().any((f) => f.path.endsWith("pubspec.yaml")) == false &&
            dir.path != "/") dir = dir.parent;
  } catch (e) {
    throw new Exception(
        "Error while trying to find pubspec.yaml from ${dir.path}:\n$e");
  }
  return dir;
}

String getPackageRootPath() => getPackageRoot().path + path.separator;

String getDifferentLines(String contents1, String contents2) {
  Set<String> str1 = new Set<String>.from(contents1.split('\n')),
      str2 = new Set<String>.from(contents2.split('\n')),
      ret,
      substract;
  ret = (str1.length > str2.length) ? str1 : str2;
  substract = (ret == str1) ? str2 : str1;
  ret.removeAll(substract);
  return ret.join('\n');
}

Future<String> processMustacheDemandingIfAbsent(
    String source, VariablesResolver resolver,
    [Type type, String message, Function constraintsChecker]) async {
  Template template = new Template(source, partialResolver: partialsResolver);
  mustacheVars(source)
      .where((k) => !resolver.getAll.containsKey(k))
      .forEach((missingVar) async {
    resolver[missingVar] = await resolver.demand(missingVar,
        type: type, message: message, constraintsChecker: constraintsChecker);
  });
  Map values = resolver.getAll;
  values.addAll(mustache_recase.cases);
  return template.renderString(values);
}

String processMustacheThrowingIfAbsent(String source, Map resolverVars) {
  Template template = new Template(source, partialResolver: partialsResolver);
  String missing;
  Map vars = new Map.from(resolverVars);
  vars.addAll(mustache_recase.cases);
  try {
    missing = mustacheVars(source).firstWhere((k) => !vars.containsKey(k));
  } on StateError catch (e) {
    return template.renderString(vars);
  }
  throw new MissingVariableException(missing);
}

class MissingVariableException {
  String name;
  MissingVariableException(this.name);
}

List<String> mustacheVars(String source) {
  Template template = new Template(source, partialResolver: partialsResolver);
  Map varsMap = gatherTemplateRequiredVars(template);
  return varsMap.keys;
}

Map gatherTemplateRequiredVars(Template template, [Map variables]) {
  Map vars = variables ?? {};
  RegExp nameRegExp = new RegExp(r": (.*).$");
  while (true) {
    TemplateException error = _failing_gathering(template, vars);
    if (error == null)
      return vars;
    else {
      String e = error.message;
      String name = nameRegExp.firstMatch(e).group(1);
      if (e.contains("for variable tag")) {
        vars[name] = "#ValueOf$name#";
      } else {
        //up to this version, if not a variable, only a Section is possible
        RegExp inSectionSrc =
            new RegExp("{{([#^])$name}}([\\s\\S]*?){{/$name}}");
        List<Match> matches = inSectionSrc.allMatches(error.source).toList();
        for (int i = 0; i < matches.length; i++) {
          String type = matches[i].group(1);
          String contents = matches[i].group(2);
          Template sectionSourceTemplate =
              new Template(contents, partialResolver: partialsResolver);
          // if (e.contains("for inverse section")) {
          // } else if (e.contains("for section")) {
          if (type == '^') {
            //inverse section
            vars["^$name"] ??= {};
            vars["^$name"]
                .addAll(gatherTemplateRequiredVars(sectionSourceTemplate));
          } else {
            vars[name] ??= {};
            vars[name]
                .addAll(gatherTemplateRequiredVars(sectionSourceTemplate));
          }
        }
      }
    }
  }
}

TemplateException _failing_gathering(Template template, Map vars,
    {bool printMessage = false, bool printReturn = false}) {
  try {
    vars.addAll(mustache_recase.cases);
    String ret = template.renderString(vars);
    if (printReturn) print(ret);
    return null;
  } on TemplateException catch (e) {
    if (printMessage) print(e.message);
    return e;
  }
}
