import 'dart:async';

import '../common.dart';
import '../generators/base.dart';
import '../generators/utils/sequencer.dart';
import '../generators/utils/variables_resolver.dart';
import '../interactions/base.dart';

import 'package:logging/logging.dart';
import 'package:path/path.dart' as path;

/// The internal base class of any operation that will be performed in the
/// workflow of the [Generator]s.
/// `T` should be the Type of the generated thing of this module
abstract class GenerationModule<T> extends GenerationStep<GenerationResult<T>> {
  VariablesResolver _resolver;
  String _pathDestination;
  Logger _logger;
  bool _overrideConfigured = false;
  OverridingPolicy _override;
  InteractionsHandler interactionsHandler;

  OverridingPolicy get override => _override;

  void set override(OverridingPolicy override) {
    _override = override;
    _overrideConfigured = true;
  }

  /// The module's execution implementation.
  /// `askIfOverridingAnswer` will be automatically provided in a re-execution
  /// if a [MightAskIfOverridingException] is thrown (user will be prompted
  /// with it's `message` property)
  /// This could also throw a [MissingVariableException] whenever this
  /// [VariableResolver]'s instance lacks the indicated value. The normal
  /// behaviour in that case should be in demanding it and re-executing this
  /// function; that's achieved by the `execute()` of the [GenerationStep]
  /// father class
  GenerationResult<T> execution(
      {bool askIfOverridingAnswer, String relativePathResolved});

  /// The [Logger] used to write the logs
  Logger get logger => _logger;

  /// The [VariablesResolver] used to get the variables for the generation
  VariablesResolver get resolver => _resolver;

  /// This package's absolute path in which to do the generation
  String get pathDestination => _pathDestination;

  /// Normalizes and absolutizes a relative to the package path and sets it
  GenerationModule(String relativePath, [OverridingPolicy overridingPolicy]) {
    String norm = path.normalize(relativePath);
    String thisPkg = getPackageRootPath();
    String abs = path.normalize(path.join(thisPkg, norm));
    if (!abs.startsWith(getPackageRootPath()))
      throw new Exception(
          "$relativePath goes outside of current package ($thisPkg)");
    this._pathDestination = abs;
    if (overridingPolicy != null) {
      this.override = overridingPolicy;
    }
  }

  /// Declare the names of the variables required for execution
  List<String> get neededVariables;
}

/// An exception intended to inform the [GeneratorModule]'s `execute()` method
/// that `execution()` code needs to know wether to override (this is a way to
/// allow the `excecution()` code to be synchronous)
class MightAskIfOverridingException implements Exception {
  final String message;
  MightAskIfOverridingException(this.message);
}

// /// An exception intended to inform the [GeneratorModule]'s `execute()` method
// /// that `execution()` code needs to know wether to override (this is a way to
// /// allow the `excecution()` code to be synchronous)
// class UnresolvedUriException implements Exception {
//   final Uri uri;
//   UnresolvedUriException(String uriStr) : this.uri = Uri.parse(uriStr);
// }

/// Internal class used to provide [GenerationModule]s with [Logger] and
/// [VariablesResolver]
class GeneratorModulesInitializer {
  VariablesResolver resolver;
  Logger logger;
  OverridingPolicy overridingDefaultValue;
  InteractionsHandler interactionsHandler;

  GeneratorModulesInitializer(this.resolver, this.logger,
      this.overridingDefaultValue, this.interactionsHandler);

  void initialize(GenerationModule module) {
    module._logger = this.logger;
    module._resolver = this.resolver;
    if (!module._overrideConfigured) module._override = overridingDefaultValue;
    module.interactionsHandler = this.interactionsHandler;
  }
}

abstract class GenerationResult<T> {
  T object;
  GenerationResult(this.object);

  bool get ofInterestForInfoLogging;
}
