import 'dart:io';
import './base.dart';
import '../common.dart';

import 'package:generator_plus/src/generators/base.dart';
import 'package:mustache/mustache.dart';
import 'package:path/path.dart' as path;
import 'package:stack_trace/stack_trace.dart';

/// Creates or copies (with processing if neccesary) a [File] when executed.
class FileGenerationModule extends GenerationModule<File> {
  /// The contents to be written in the generated [File]
  final String sourceString;

  /// Wether the generation should process the mustache code or copy it "as is"
  bool processInputWithMustache;

  /// This will create a [File] in the provided `relativePathDestination`
  /// (relative to this package location) by copying (or, eventually, processing)
  /// the provided `relativePathDestination` (being the processing determied by
  /// the `processInputWithMustache` flag that defaults to `true`)
  FileGenerationModule(String contents, String relativePathDestination,
      {this.processInputWithMustache: true})
      : super(relativePathDestination),
        this.sourceString = contents;

  /// If the file to copy is named `name.extension.mustache` the generated [File]
  /// will be the result of the processing of the source one into `name.extension`
  /// But if source has only one extension, even `name.mustache`, the generated
  /// [File] will, by default, not be processed and named as the source
  FileGenerationModule.fromExistingFile(
      File source, String relativePathDestination,
      {bool processInputWithMustache: true})
      : this(source.readAsStringSync(), relativePathDestination,
            processInputWithMustache: processInputWithMustache);

  FileGenerationModule.fromExistingFilePath(
      String sourcePath, String relativePathDestination,
      {bool processInputWithMustache: true})
      : this(new File(sourcePath).readAsStringSync(), relativePathDestination,
            processInputWithMustache: processInputWithMustache);

  FileGenerationModule.fromExistingDir(
      Directory dir, String generationName, String sourceString,
      {bool processInputWithMustache: true})
      : this(
            sourceString,
            path.relative(path.join(dir.path, generationName),
                from: getPackageRootPath()),
            processInputWithMustache: processInputWithMustache);

  @override
  List<String> get neededVariables => mustacheVars(sourceString);
  @override
  FileGenerationResult execution(
      {bool askIfOverridingAnswer, String relativePathResolved}) {
    File destinationFile = new File(
        processMustacheThrowingIfAbsent(this.pathDestination, resolver.getAll));
    bool overriden = false;
    logger.finest(
        "destination File instance created: <File>(${destinationFile.path})");
    if (destinationFile.existsSync()) {
      String alreadyExistsMessage =
          "File '${destinationFile.path}' already exists";
      logger.warning(alreadyExistsMessage);
      switch (this.override) {
        case OverridingPolicy.ALWAYS:
          logger.fine("Override policy is ALWAYS, file is being overriden");
          overriden = true;
          break;
        case OverridingPolicy.ASK:
          logger.fine("Override policy is ASK");
          if (askIfOverridingAnswer == null)
            throw new MightAskIfOverridingException(
                "$alreadyExistsMessage. Override it?");
          if (askIfOverridingAnswer)
            overriden = true;
          else
            return new FileGenerationResult(
                destinationFile, false, processInputWithMustache, false);
          break;
        case OverridingPolicy.NEVER:
          logger.fine("Override policy is NEVER");
          String e = "$alreadyExistsMessage. Overriding isn't allowed.";
          logger.severe(e);
          throw new StateError(e);
      }
    } else {
      logger.finer(
          "File ${destinationFile.path} doesn't exists. Will create it...");
      // destinationFile.createSync();
    }
    logger.finest("Processing mustache: $processInputWithMustache");
    String output = processInputWithMustache
        ? processMustacheThrowingIfAbsent(sourceString, resolver.getAll)
        : sourceString;
    bool processed = sourceString != output;
    logger.finer("Writing to file (${destinationFile.path})...");
    destinationFile.writeAsStringSync(output);
    logger.finest("${destinationFile.path} written.");
    return new FileGenerationResult(
        destinationFile, overriden, processInputWithMustache, processed);
  }
}

class FileGenerationResult extends GenerationResult<File> {
  bool overriden;
  bool processed;
  bool shouldMustacheBeProcessed;
  int lenght;

  FileGenerationResult(File object, this.overriden,
      this.shouldMustacheBeProcessed, this.processed)
      : super(object) {
    this.lenght = this.object.lengthSync();
  }

  @override
  bool get ofInterestForInfoLogging => true;
}
