library generator_plus.base;

import 'dart:async';
import 'package:generator_plus/src/generators/utils/variables_resolver.dart';
import 'package:logging/logging.dart';

import "generation/fileProcessorAnnotations/base.dart";
import './generators/base.dart' show Generator, GenerationResults;
import './interactions/base.dart' show IOInterface, InteractionsHandler;

/// The execution starting point entity. All what Generator+ would do
/// should be defined here.
/// This is meant to be the base class of a generated class that would be
/// present in `<package_root>/tool/generator+/generator+.dart` in order to
/// separate the execution logic of this package from the interfaces that will
/// execute them, giving an uniform way of execution to all of them.
abstract class GeneratorPlus {
  List<Generator> _generators = [];

  /// The global generators that would be available by default to everything
  /// that uses this package
  List<Generator> get generators => _generators;

  /// This is intended to extend the global `generators` with the
  /// specific-to-the-package ones
  void addGenerators(List<Generator> generators) =>
      this._generators.addAll(generators);

  /// Should be called by the [GeneratorPlusRunnerAPI] in order to provide
  /// the tools needed for a right execution in the stated interface
  void _initialize(IOInterface iface, InteractionsHandler interactionsHandler,
      {Logger logger, VariablesResolver varsResolver}) {
    generators.forEach((Generator g) {
      g.initialize(interactionsHandler,
          logger: logger, varsResolver: varsResolver);
    });
  }

  /// That what gives sense to this class
  Future<GenerationResults> execute(Generator generator) {
    if (this.generators.contains(generator)) {
      return generator.execute();
    } else
      throw new Exception("${generator.name} not declared in this Generator+");
  }
}

/// The interface used to execute [GeneratorPlus] in several environments.
/// This is intended to implement the vissicitudes of every platform that will
/// use this library (i.e. CLI, IDEs, etc.) providing the same functionality on
/// each of them
abstract class GeneratorPlusRunnerAPI {
  IOInterface iface;
  GeneratorPlus implementation;
  InteractionsHandler interactionsHandler;
  Logger logger;
  VariablesResolver varsResolver;

  GeneratorPlusRunnerAPI(this.implementation, this.iface,
      this.interactionsHandler, this.logger, this.varsResolver) {
    if (!iface.setUp())
      throw new Exception("Interface couldn't be initialized (setUp fail)");
    implementation._initialize(iface, interactionsHandler,
        logger: logger, varsResolver: varsResolver);
  }

  Future run(List<String> args) {
    Completer completer = new Completer();
    logger.finer(
        "Starting execution of ${implementation.runtimeType} with $args");
    execution(args).then((_) {
      logger.finer(
          "Finished execution of ${implementation.runtimeType} seccessfully");
      if (!iface.tearDown())
        logger.warning("Coudln't re-stablish terminal line/echo mode");
      completer.complete();
    }).catchError((e) {
      logger.finer(
          "Finished execution of ${implementation.runtimeType} erroneously");
      iface.tearDown();
      completer.completeError(e);
    });
    return completer.future;
  }

  /// The main execution logic that will run the [GeneratorPlus]
  Future execution(List<String> args);

  // /// Will show the user the available generators that he could use
  // listGenerators();
}
