import 'dart:developer';
import "dart:io";
import "dart:async";
// import 'package:generator_plus/src/generators/utils/template_utils.dart';

import 'package:generator_plus/src/common.dart';
import 'package:generator_plus/src/generators/utils/template_utils.dart';

import '../generation/base.dart';
import '../interactions/base.dart';
import '../generation/dependency.dart';
import './utils/sequencer.dart';
import './utils/variables_resolver.dart';
import './utils/interaction_module.dart';
import '../generation/base.dart';
import '../generation/file_processor.dart';
import '../generation/new_dir.dart';
import '../generation/new_file.dart';
import 'package:generator_plus/src/generators/utils/template_utils.dart';
import 'package:logging/logging.dart';

/// The base class from where the generation will took place. This is the
/// backbone of the generation and is the starting point from where new
/// [Generator]s shall start.
abstract class Generator implements InteractionsInterface {
  DependenciesProcessor _depsProcessor;
  GeneratorModulesInitializer _modulesInitializer;
  bool _predefinedDependenciesAdded = false;
  bool _initialized = false;
  GenerationStepsSequencer _sequencer;
  InteractionsHandler interactionsHandler;
  VariablesResolver variablesResolver;
  GenerationResults generationResults;
  Logger logger;

  Generator() {
    _depsProcessor = new DependenciesProcessor();
    _sequencer = new GenerationStepsSequencer();
  }

  /// Here are intended to be provided all the tools that will be used in the
  /// generation. This should be called before execution from the
  Future initialize(InteractionsHandler interactionsHandler,
      {Logger logger, VariablesResolver varsResolver}) {
    if (_initialized) throw new Exception("This was already initialized");
    this.interactionsHandler = interactionsHandler;
    this.variablesResolver = varsResolver ?? new VariablesResolver({});
    this.variablesResolver.missingResolver ??= (DemandingVariable dv) async {
      String ret;
      do {
        ret = await askForInput(dv.message, null);
      } while (!dv.constraintsChecker(ret));
      return ret;
    };
    if (this.startingVariables != null)
      this.variablesResolver.addAll(this.startingVariables);
    this.logger =
        logger ?? new Logger("generator." + this.runtimeType.toString());
    _modulesInitializer = new GeneratorModulesInitializer(
        this.variablesResolver,
        this.logger,
        this.overridePolicy,
        this.interactionsHandler);
    _initialized = true;
  }

  /// The execution logic of this generator's generation
  Future execution();

  /// Here you can define starting variables for the generation
  Map get startingVariables;

  /// The name of this [Generator]
  String get name;

  /// A description for this [Generator]
  String get description;

  /// The, by default, `override` value this [Generator]'s [GeneratorModule]
  /// will have
  OverridingPolicy get overridePolicy;

  /// Define the dependencies this generator needs in order to work properly,
  /// they will be automatically included or updated in the local 'pubspec.yaml'
  List<Dependency> get alwaysNeededDependencies;

  Set<Dependency> get dependencies => _depsProcessor.deps;

  /// Returns true if `dependency` addition was successful and didn't existed before
  bool addDependency(Dependency dependency) {
    _addPredefinedDeps();
    bool ret = _depsProcessor.addDependency(dependency);
    //Split message to prevent warnings
    String logMsg = "Adding dependency '${dependency.name}' ";
    logMsg += ret ? "succeeded" : "failed";
    logger.finer(logMsg);
    return ret;
  }

  /// The principal bone adder, in the very backbone of this [Generator]
  /// (I feel like a poet XD)
  Future addGenerationStep(GenerationStep step) {
    logger.finer("Adding a ${step.runtimeType} GenerationStep");
    return _sequencer.addStep(step);
  }

  /// Syntactic-sugar for standard processing of a file. _Id est_ interacting
  /// with all its annotatons
  Future<FileProcessResult> processFile(String relativePath,
      {Map<String, String> templates,
      List generationIds: const [],
      List generationIdsExcluded: const []}) {
    FileProcessorModule fileProcessor = new FileProcessorModule(relativePath,
        templates: templates,
        generationIds: generationIds,
        generationIdsExcluded: generationIdsExcluded);
    return addGenerationStep(fileProcessor);
  }

  /// Syntactic-sugar for a new [File] creation
  Future<FileGenerationResult> createFile(
      String contents, String relativePathDestination) {
    FileGenerationModule fileGeneration =
        new FileGenerationModule(contents, relativePathDestination);
    return addGenerationStep(fileGeneration);
  }

  /// Syntactic-sugar for a new [Directory] creation
  Future<DirGenerationResult> createDir(String relativePath) {
    DirGenerationModule dirGeneration = new DirGenerationModule(relativePath);
    return addGenerationStep(dirGeneration);
  }

  Future generateTemplateDir(String sourceRelativePath) async {
    // FileGenerationModule module = new FileGenerationModule.fromExistingFilePath(
    //     "lib/src/templates/generators/tool/generator+/generators/{{name}}.dart.mustache",
    //     "tool/generator+/generators/{{name}}.dart");
    logger.finer("generating template dir $sourceRelativePath");
    Directory sourceDir = await sourceTemplateDir(sourceRelativePath);
    List<GenerationModule> modules = templateGenerationModules(sourceDir);
    if (modules.length > 1)
      for (int i = 0; i < modules.length; i++) addGenerationStep(modules[i]);
    logger.finer(
        "Template dir generation of $sourceRelativePath processed with ${modules.length} modules loaded");
  }

  void _addPredefinedDeps() {
    if (_predefinedDependenciesAdded == false &&
        alwaysNeededDependencies != null &&
        alwaysNeededDependencies.isNotEmpty) {
      _depsProcessor.addDependencies(alwaysNeededDependencies);
      _predefinedDependenciesAdded = true;
    }
  }

  @override
  Future<bool> askForConfirmation(String message,
      {bool defaultValue: true, String varName}) {
    ConfirmationInteractionModule confirmationInteractionModule =
        new ConfirmationInteractionModule(interactionsHandler, message,
            defaultValue: defaultValue);
    return addGenerationStep(confirmationInteractionModule);
    // return confirmationInteractionModule.returnFuture;
  }

  @override
  Future<String> askForInput(String message, String checkRegExp,
      {String varName}) {
    InputInteractionModule inputInteractionModule =
        new InputInteractionModule(interactionsHandler, message, checkRegExp);
    return addGenerationStep(inputInteractionModule);
    // return inputInteractionModule.returnFuture;
  }

  @override
  Future<String> askForSelection(String message, List<String> options,
      {String varName}) {
    SelectionInteractionModule selectionInteractionModule =
        new SelectionInteractionModule(interactionsHandler, message, options);
    return addGenerationStep(selectionInteractionModule);
    // return selectionInteractionModule.returnFuture;
  }

  bool _runPubGetDependencies;
  Completer<GenerationResults> _executionCompleter;
  Future<GenerationResults> execute({bool runPubGetDependencies: false}) {
    _executionCompleter = new Completer();
    new Future(() async {
      _runPubGetDependencies = runPubGetDependencies;
      if (!_initialized)
        throw new Exception("must initialize() before execution");
      await execution();
      GenerationStep preLastStep = new NoOpGenerationStep();
      GenerationStep lastStep = new NoOpGenerationStep();
      _sequencer.addStep(preLastStep).then(_preExecutionFinish);
      _sequencer.addStep(lastStep).then(_executionFinish);
      _addPredefinedDeps();
      await _sequencer.initializeSteps(_modulesInitializer);
      generationResults = new GenerationResults();
      _sequencer.generationStepResults.listen((GenerationResult result) {
        if (result is FileGenerationResult)
          generationResults.filesGenerated.add(result);
        else if (result is DirGenerationResult)
          generationResults.dirsGenerated.add(result);
        else if (result is FileProcessResult)
          generationResults.filesProcessed.add(result);
      });
      _sequencer.execute().catchError((e) {
        _executionCompleter.completeError(e);
      });
    });
    return _executionCompleter.future;
  }

  _preExecutionFinish(value) async {
    logger.finer("Searching for editor executable.");
    if (generationResults.filesGenerated.isNotEmpty) {
      logger.fine("Opening generated files in 'atom' editor.");
      generationResults.filesGenerated.forEach((FileGenerationResult fgr) {
        Process.start("atom", [fgr.object.path],
            mode: ProcessStartMode.DETACHED);
      });
    }

    generationResults.newPubspecLines = await _depsProcessor.processPubspec();
    if (_runPubGetDependencies) {
      logger.info("Executing 'pub get'");
      Process.start('pub', ['get'],
          mode: ProcessStartMode.DETACHED,
          workingDirectory: getPackageRootPath());
    } else if (_depsProcessor.status == DependencyProcessorStatus.CHANGED) {
      logger.finest("Asking if running 'pub get' or not");
      askForConfirmation("Dependencies changed, run `pub get`?").then((bool c) {
        if (c) {
          Process.start('pub', ['get'],
              mode: ProcessStartMode.DETACHED,
              workingDirectory: getPackageRootPath());
        }
      });
    }
    return;
  }

  _executionFinish(_) async {
    logger.fine("Execution of ${this.runtimeType} finished.");
    _executionCompleter.complete(generationResults);
  }
}

enum OverridingPolicy { NEVER, ASK, ALWAYS }

/// The class summarizing all the changes the execution made
//TODO: meter todo lo que falta
class GenerationResults {
  String newPubspecLines;
  ProcessResult pubGetResult;
  List<FileProcessResult> filesProcessed = [];
  List<FileGenerationResult> filesGenerated = [];
  List<DirGenerationResult> dirsGenerated = [];
}
