import 'dart:async';
// import 'dart:isolate';
import '../../generation/base.dart';
import '../../common.dart';
import '../base.dart';
import './interaction_module.dart';
import './template_utils.dart';
import 'package:logging/logging.dart';

/// Internal class that holds the execution logic of the [Generator], it will
/// handle the execution order of the [GenerationModule]s & [InteractionModule]s
///
/// It's intenteded to stack [GenerationStep]s that will be ran in the order
/// they are stacked. But, if their execution stacks another [GenerationStep],
/// it will be executed next to the one that stacked it
class GenerationStepsSequencer {
  List<GenerationStep> _steps = [];
  int _currentIndex, _duringExecutionAddedIndex;
  bool _executionStarted = false;
  bool _stepsInitialized = false;
  Completer _completer;
  Logger _logger;

  GeneratorModulesInitializer _modulesInitializer;
  InteractionModuleInitializer _interactionModuleInitializer;

  StreamController<GenerationResult> _resultsStream;
  Stream get generationStepResults {
    _resultsStream = new StreamController();
    return _resultsStream.stream;
  }

  TrackingFuture addStep(GenerationStep step) {
    if (_stepsInitialized) _initializeStep(step);
    if (!_executionStarted) {
      _steps.add(step);
      _logger?.finest(
          "Sequencer addStep(): Now there are ${_steps.length} steps (execution not started yet)");
    } else {
      _duringExecutionAddedIndex ??= _currentIndex;
      _steps.insert(++_duringExecutionAddedIndex, step);
      _logger?.finest(
          "Sequencer addStep(): While executing $_currentIndexº step, added a new step on position $_duringExecutionAddedIndex/${_steps.length} (execution started)");
    }
    return step._future;
  }

  Future execute() {
    _completer = new Completer();
    _currentIndex = -1; //_executeNextStep() increments it inmediatly
    _executionStarted = true;
    new Future.microtask(() {
      _executeNextStep();
    });
    return _completer.future;
  }

  _executeNextStep([GenerationResult executionResult]) {
    if (executionResult != null) _resultsStream.add(executionResult);
    if (!_completer.isCompleted) {
      _duringExecutionAddedIndex = null;
      if (_currentIndex < _steps.length - 1) {
        _currentIndex++;
        GenerationStep generationStep = _steps[_currentIndex];
        TrackingFuture stepFuture = generationStep._future;
        if (stepFuture.isAwaited) {
          stepFuture.thenExecutionAfter
              .then(_executeNextStep, onError: _endWithError);
        } else {
          stepFuture.then(_executeNextStep, onError: _endWithError);
        }
        generationStep.execute();
      } else
        _completer.complete();
    }
  }

  _endWithError(e) {
    _completer.completeError(e);
  }

  initializeSteps(GeneratorModulesInitializer modulesInitializer) {
    _modulesInitializer = modulesInitializer;
    this._logger ??= modulesInitializer.logger;
    _interactionModuleInitializer =
        new InteractionModuleInitializer(modulesInitializer.logger);
    _steps.forEach(_initializeStep);
    _stepsInitialized = true;
  }

  void _initializeStep(GenerationStep step) {
    if (step is GenerationModule)
      _modulesInitializer.initialize(step);
    else if (step is InteractionModule)
      _interactionModuleInitializer.initialize(step);
  }
}

/// Only subclasses of this class may be used by a [GenerationStepsSequencer]
/// `T` would be the Type of the result of executing this
@proxy
abstract class GenerationStep<T> {
  TrackingCompleter _completer = new TrackingCompleter<T>();
  Logger get logger;

  /// This operation must hold the operational logic of the subclass that, when
  /// executed, will complete the [Future] with the *eventually needed* value
  /// for the [GenerationStepsSequencer]
  T execution();

  TrackingFuture<T> get _future => _completer.future;

  _execute({bool answer, String relativePathResolved}) {
    logger.fine("starting execution of ${this.runtimeType}");
    var ending;
    Future retFuture;
    try {
      if (answer == null && null == relativePathResolved)
        ending = execution();
      else
        ending = (this as GenerationModule).execution(
            askIfOverridingAnswer: answer,
            relativePathResolved: relativePathResolved);
      // } on UnresolvedUriException catch (e) {
      //   Isolate.resolvePackageUri(e.uri).then((Uri uri) {
      //     _execute(relativePathResolved: uri.path);
      //   });
      //   return;
    } on MightAskIfOverridingException catch (e) {
      this
          .interactionsHandler
          .askForConfirmation(e.message)
          .then((bool answer) {
        _execute(answer: answer);
      });
      return;
    } on MissingPartialException catch (e) {
      logger.finer(
          "Needing to resolve ${e.partialName} (already tried in generator_plus package: ${e.original_package_searched})");
      if (e.original_package_searched) rethrow;
      resolvePath("package:generator_plus/src").then((String path) {
        original_package_path =
            path.substring(0, path.length - 7); //from src/common.dart
        logger.fine("mustache_partials dir resolved to $original_package_path");
        _execute();
      });
      return;
    } on MissingVariableException catch (e) {
      String recasingAttempt = this.resolver.get(e.name);
      if (recasingAttempt != null) {
        this.resolver[e.name] = recasingAttempt;
        _execute();
        return;
      } else {
        this
            .interactionsHandler
            .askForInput(
                "Please provide the value for the missing variable '${e.name}'",
                null)
            .then((String value) {
          this.resolver[e.name] = value;
          _execute();
        });
        return;
      }
    } catch (e, st) {
      _completer.completeError(e, st);
      return;
    }
    try {
      retFuture = this.returnFuture;
    } on NoSuchMethodError catch (e) {
      retFuture = (ending is Future) ? ending : new Future.value(ending);
    }
    retFuture.then(_execute_ending, onError: _execute_error_ending);
  }

  /// The function wrapping this step execution pretended to be called only by
  /// this library
  TrackingFuture<T> execute() {
    new Future.microtask(() {
      _execute();
    });
    return _future;
  }

  _execute_error_ending(e) {
    logger.severe("ending execution of ${this.runtimeType} erroneously");
    _completer.completeError(e);
  }

  _execute_ending(stepResult) {
    if (stepResult is GenerationResult) {
      if (stepResult.ofInterestForInfoLogging) logger.info(stepResult);
    } else if (stepResult != null) logger.info(stepResult);
    logger.finest("ending execution of ${this.runtimeType}");
    _completer.complete(stepResult);
  }
}

/// A [GenerationStep] that doesn't do anything (No-Operation)
class NoOpGenerationStep extends GenerationStep {
  execution() {}

  Logger get logger => new Logger('');
}

///A [Completer] that creates hacked [Future]s: [TrackingFuture]
class TrackingCompleter<T> implements Completer<T> {
  Completer _internal;
  TrackingFuture _trackingFuture;

  TrackingCompleter() {
    _internal = new Completer();
    _trackingFuture = new TrackingFuture(_internal.future);
  }

  void complete([value]) {
    _internal.complete(value);
  }

  void completeError(Object error, [StackTrace stackTrace]) {
    if (stackTrace != null)
      _internal.completeError(error, stackTrace);
    else
      _internal.completeError(error);
  }

  TrackingFuture get future => _trackingFuture;

  bool get isCompleted => _internal.isCompleted;
}

///A [Future] that is hacked for following it's execution life status
class TrackingFuture<T> implements Future<T> {
  Future _internal;

  /// Wether if it's `then` has been called, in order to know if it code is
  /// queued for execution after this finish
  bool isAwaited = false;

  /// Wether if code quedued for execution after completion has been ran or not
  bool isThenExecuted = false;

  /// Wether if code quedued for execution after completion is being run or not
  bool isThenBeingExecuted = false;

  Function _trackedThenFunction;

  Completer _completerBefore = new Completer();
  Completer _completerAfter = new Completer();

  TrackingFuture(this._internal);

  @override
  Stream asStream() => _internal.asStream();

  @override
  Future catchError(Function onError, {bool test(Object error)}) {
    if (test != null)
      return _internal.catchError(onError, test: test);
    else
      return _internal.catchError(onError);
  }

  @override
  Future then<S>(onValue(value), {Function onError}) {
    if (isAwaited == true)
      throw new Exception("This future has been quedued more than once");
    isAwaited = true;
    _trackedThenFunction = onValue;
    if (onError != null)
      _internal.then(_trackedThenBegin, onError: onError);
    else
      _internal.then(_trackedThenBegin);
    return _completerBefore.future;
  }

  @override
  Future timeout(Duration timeLimit, {onTimeout()}) =>
      _internal.timeout(timeLimit, onTimeout: onTimeout);

  @override
  Future whenComplete(action()) => _internal.whenComplete(action);

  Function _trackedThenBegin(value) {
    if (isThenExecuted == true)
      throw new Exception("This future's then() has been ran more than once");
    isThenBeingExecuted = true;
    _completerBefore.complete(null);
    var ret = _trackedThenFunction(value);
    if (ret is Future)
      ret.then(_trackedThenEnd);
    else
      _trackedThenEnd(value);
  }

  Function _trackedThenEnd(value) {
    isThenBeingExecuted = false;
    isThenExecuted = true;
    _completerAfter.complete(null);
  }

  Future get thenExecutionBefore {
    if (!isAwaited)
      throw new Exception("This haven't been queued in a .then()");
    return _completerBefore.future;
  }

  Future get thenExecutionAfter {
    if (!isAwaited)
      throw new Exception("This haven't been queued in a .then()");
    return _completerAfter.future;
  }
}
