import 'dart:async';
import 'dart:developer';
import 'dart:mirrors';

import 'package:recase/recase.dart';

const Map<String, String> recasingsShorthands = const {
  "cc": "camelCase",
  "sc": "snakeCase",
  "pc": "pascalCase",
};

/// Will hold a bunch of variables and provide tools for obtaining them
class VariablesResolver {
  Map<String, Variable> _mem;
  Function missingResolver;

  VariablesResolver([Map initialVars, Function demandingMissingresolver]) {
    this._mem = initialVars ?? new Map();
    this.missingResolver = demandingMissingresolver;
  }

  /// Returns the requested variable, or `null` if it's not defined
  get(String request) {
    RegExp casingAnalyzer = new RegExp(r"([A-Za-z0-9]+)_?([A-Za-z]*)?");
    Match match = casingAnalyzer.firstMatch(request);
    String varName = match.group(1);
    String recase = match.group(2);
    if (recase == null)
      return _mem[varName];
    else {
      if (recasingsShorthands.keys.contains(recase))
        recase = recasingsShorthands[recase];
      return _mem[varName][recase];
    }
  }
  // get(String varName) => _mem[varName] ?? null;
  // get(String varName) => _mem.containsKey(varName) ? _mem[varName] : null;

  Map get getAll => new Map.unmodifiable(_mem);

  /// Returns a Future that will be completed with the demanding `varName`'s
  /// value as soon as that's available
  Future demand(String varName,
      {Type type: dynamic, String message, Function constraintsChecker}) {
    Completer completer = new Completer();
    new Future(() {
      if (_mem[varName] == null) {
        DemandingVariable demandingVariable =
            new DemandingVariable(varName, type: type);
        if (message == null) {
          demandingVariable.message =
              "please provide the variable value for '$varName'";
        } else {
          demandingVariable.message = message;
        }
        if (constraintsChecker != null)
          demandingVariable.constraintsChecker = constraintsChecker;
        demandingVariable.type = type;
        debugger();
        var ret = missingResolver(demandingVariable);
        if (ret is Future)
          ret.then((resolvedValue) {
            completer.complete(resolvedValue);
          });
        else
          completer.complete(ret);
      } else {
        completer.complete(_mem[varName]);
      }
    });
    return completer.future;
  }

  Variable operator [](key) => get(key);

  void operator []=(key, value) {
    if (key is! String) throw new MalformedVariablesMap(key, value);
    if (value is String)
      _mem[key] = new Variable(value);
    else if (value is Variable)
      _mem[key] = value;
    else
      throw new MalformedVariablesMap(key, value);
  }

  void addAll(Map vars) {
    Map<String, Variable> variables = {};
    vars.forEach((key, value) {
      if (key is! String) throw new MalformedVariablesMap(key, value);
      if (value is String)
        variables[key] = new Variable(value);
      else if (value is Variable)
        variables[key] = value;
      else
        throw new MalformedVariablesMap(key, value);
    });
    _mem.addAll(variables);
  }
}

class MalformedVariablesMap implements Exception {
  String message;
  MalformedVariablesMap(key, value)
      : this.message =
            "Map<String,[String|Variable]> must be provided. One of the tuples is a <${key.runtimeType}, ${value.runtimeType}>";
}

class Variable {
  String _original;
  ReCase _reCasedOriginal;

  Variable(this._original) {
    _reCasedOriginal = new ReCase(_original);
  }

  String toString() => _original;

  String operator [](key) => _mirror(key);

  // //ReCase implementation
  // String get camelCase => _reCasedOriginal.camelCase;
  // String get constantCase => _reCasedOriginal.constantCase;
  // String get dotCase => _reCasedOriginal.dotCase;
  // String get headerCase => _reCasedOriginal.headerCase;
  // String get paramCase => _reCasedOriginal.paramCase;
  // String get pascalCase => _reCasedOriginal.pascalCase;
  // String get pathCase => _reCasedOriginal.pathCase;
  // String get sentenceCase => _reCasedOriginal.sentenceCase;
  // String get snakeCase => _reCasedOriginal.snakeCase;
  // String get titleCase => _reCasedOriginal.titleCase;

  String _mirror(key) {
    InstanceMirror mirror = reflect(_reCasedOriginal);
    Symbol symbol = new Symbol(key);
    InstanceMirror invokedResult = mirror.getField(symbol);
    return invokedResult.reflectee;
  }
}

class DemandingVariable {
  String name;
  Type type;
  Function constraintsChecker;
  String message;
  DemandingVariable(this.name,
      {this.type, this.message, this.constraintsChecker});
}
