import 'package:generator_plus/generator_plus.dart';
import 'package:generator_plus/src/generation/base.dart';
import 'package:generator_plus/src/generators/utils/sequencer.dart';
import 'package:generator_plus/src/generators/utils/variables_resolver.dart';
import 'package:logging/logging.dart';
import 'package:test/test.dart';
import 'dart:async';

main() => defineTests();

class Printer extends GenerationStep<int> {
  int index;
  Printer(this.index);
  int execution() {
    print("PRINTING: $index");
    printerInts.add(this.index);
    return this.index;
  }

  @override
  Logger get logger => new Logger("logger");
}

List<int> printerInts = [];

defineTests() {
  group('Sequencer', () {
    VariablesResolver resolver = new VariablesResolver();
    Logger logger = new Logger("srp");
    OverridingPolicy overridingDefaultValue = OverridingPolicy.NEVER;
    GeneratorModulesInitializer initializer = new GeneratorModulesInitializer(
        resolver, logger, overridingDefaultValue, null);
    test('runs in correct order its steps', () async {
      GenerationStepsSequencer sequencer = new GenerationStepsSequencer();
      sequencer.addStep(new Printer(0));
      sequencer.addStep(new Printer(1)).then((int index) async {
        if (index == 1) {
          sequencer.addStep(new Printer(2)).then((int index) async {
            sequencer.addStep(new Printer(3));
          });
          sequencer.addStep(new Printer(4)).then((int index) async {
            sequencer.addStep(new Printer(5));
            sequencer.addStep(new Printer(6));
          });
        }
      });
      sequencer.addStep(new Printer(7)).then((int index) {});
      Completer completer = new Completer();
      sequencer.initializeSteps(initializer);
      sequencer.execute().then((_) {
        completer.complete(printerInts);
      });
      expect(completer.future, completion(equals([0, 1, 2, 3, 4, 5, 6, 7])));
    });
    test('runs in correct order its steps (with delays in between)', () async {
      GenerationStepsSequencer sequencer = new GenerationStepsSequencer();
      printerInts = [];
      sequencer.addStep(new Printer(0));
      sequencer.addStep(new Printer(1)).then((int index) async {
        if (index == 1) {
          await new Future.delayed(new Duration(milliseconds: 90));
          sequencer.addStep(new Printer(2)).then((int index) async {
            await new Future.delayed(new Duration(milliseconds: 90));
            sequencer.addStep(new Printer(3));
          });
          await new Future.delayed(new Duration(milliseconds: 90));
          sequencer.addStep(new Printer(4)).then((int index) async {
            sequencer.addStep(new Printer(5));
            await new Future.delayed(new Duration(milliseconds: 90));
            sequencer.addStep(new Printer(6));
          });
        }
      });
      sequencer.addStep(new Printer(7));
      Completer completer = new Completer();
      sequencer.initializeSteps(initializer);
      sequencer.execute().then((_) {
        completer.complete(printerInts);
      });
      expect(completer.future, completion(equals([0, 1, 2, 3, 4, 5, 6, 7])));
    });
    test("runs well when first executed being empty", () {
      GenerationStepsSequencer sequencer = new GenerationStepsSequencer();
      printerInts = [];
      Completer completer = new Completer();
      sequencer.initializeSteps(initializer);
      sequencer.execute().then((_) {
        sequencer.addStep(new Printer(0)).then((int index) {
          sequencer.addStep(new Printer(1));
          sequencer.addStep(new Printer(2)).then((int index) {
            sequencer.addStep(new Printer(3)).then((int index) {
              sequencer.addStep(new Printer(4)).then((int index) {
                sequencer.addStep(new Printer(5));
              });
            });
          });
        });
        sequencer.addStep(new Printer(6));
        sequencer.addStep(new Printer(7)).then((_) {
          completer.complete(printerInts);
        });
      });
      new Future(() => sequencer.execute());
      expect(completer.future, completion(equals([0, 1, 2, 3, 4, 5, 6, 7])));
    });
  });
  group("Tracking Futures:", () {
    test("isAwaited", () {
      TrackingCompleter completer = new TrackingCompleter();
      TrackingFuture future = completer.future;
      expect(future.isAwaited, isFalse);
      future.then((_) {});
      expect(future.isAwaited, isTrue);
    });
    test("isThenExecuted", () {
      TrackingCompleter completer = new TrackingCompleter();
      TrackingFuture future = completer.future;
      Completer testCompleter = new Completer();
      expect(future.isThenExecuted, isFalse);
      future.then((_) {
        expect(future.isThenExecuted, isFalse);
        new Future.delayed(new Duration(milliseconds: 50), () {
          expect(future.isThenExecuted, isTrue);
          testCompleter.complete(null);
        });
      });
      expect(future.isThenExecuted, isFalse);
      completer.complete();
      new Future(() {
        expect(future.isThenExecuted, isTrue);
      });
      expect(testCompleter.future, completes);
    });
    test("isThenBeingExecuted", () {
      TrackingCompleter completer = new TrackingCompleter();
      TrackingFuture future = completer.future;
      Completer testCompleter = new Completer();
      expect(future.isThenBeingExecuted, isFalse);
      future.then((_) {
        expect(future.isThenBeingExecuted, isTrue);
        new Future.delayed(new Duration(milliseconds: 50), () {
          expect(future.isThenBeingExecuted, isFalse);
          testCompleter.complete(null);
        });
      });
      expect(future.isThenBeingExecuted, isFalse);
      completer.complete();
      new Future(() {
        expect(future.isThenBeingExecuted, isFalse);
      });
      expect(testCompleter.future, completes);
    });
    test("all works well with await", () async {
      TrackingCompleter completer = new TrackingCompleter();
      TrackingFuture future = completer.future;
      Completer testCompleter = new Completer();
      expect(future.isAwaited, isFalse);
      expect(future.isThenExecuted, isFalse);
      new Future(() {
        completer.complete();
      });
      testCompleter.future.then((_) {
        expect(future.isThenExecuted, isTrue);
        expect(future.isThenBeingExecuted, isFalse);
      });
      var ret = await future;
      expect(future.isAwaited, isTrue);
      expect(future.isThenBeingExecuted, isTrue);
      expect(future.isThenExecuted, isFalse);
      testCompleter.complete(null);
    });
  });
}
